@@ -39,6 +39,7 @@
 #include "gfileicon.h"
 #include <glib/gstdio.h>
 #include "glibintl.h"
+#include "giomodule-priv.h"
 
 #include "gioalias.h"
 
@@ -53,13 +54,15 @@
  **/
 
 #define DEFAULT_APPLICATIONS_GROUP  "Default Applications" 
+#define ADDED_ASSOCIATIONS_GROUP    "Added Associations" 
 #define REMOVED_ASSOCIATIONS_GROUP  "Removed Associations" 
 #define MIME_CACHE_GROUP            "MIME Cache"
 
-static void g_desktop_app_info_iface_init (GAppInfoIface *iface);
-
-static GList *get_all_desktop_entries_for_mime_type (const char *base_mime_type);
-static void mime_info_cache_reload (const char *dir);
+static void     g_desktop_app_info_iface_init         (GAppInfoIface    *iface);
+static GList *  get_all_desktop_entries_for_mime_type (const char       *base_mime_type);
+static void     mime_info_cache_reload                (const char       *dir);
+static gboolean g_desktop_app_info_ensure_saved       (GDesktopAppInfo  *info,
+						       GError          **error);
 
 /**
  * GDesktopAppInfo:
@@ -164,6 +167,22 @@ g_desktop_app_info_init (GDesktopAppInfo *local)
 {
 }
 
+static char *
+binary_from_exec (const char *exec)
+{
+  const char *p, *start;
+  
+  p = exec;
+  while (*p == ' ')
+    p++;
+  start = p;
+  while (*p != ' ' && *p != 0)
+    p++;
+  
+  return g_strndup (start, p - start);
+  
+}
+
 /**
  * g_desktop_app_info_new_from_filename:
  * @filename: a string containing a file name.
@@ -217,7 +236,7 @@ g_desktop_app_info_new_from_filename (const char *filename)
                                     G_KEY_FILE_DESKTOP_GROUP,
                                     G_KEY_FILE_DESKTOP_KEY_TRY_EXEC,
                                     NULL);
-  if (try_exec)
+  if (try_exec && try_exec[0] != '\0')
     {
       char *t;
       t = g_find_program_in_path (try_exec);
@@ -264,18 +283,7 @@ g_desktop_app_info_new_from_filename (const char *filename)
     }
   
   if (info->exec)
-    {
-      char *p, *start;
-
-      p = info->exec;
-      while (*p == ' ')
-	p++;
-      start = p;
-      while (*p != ' ' && *p != 0)
-	p++;
-      
-      info->binary = g_strndup (start, p - start);
-    }
+    info->binary = binary_from_exec (info->exec);
   
   return info;
 }
@@ -381,7 +389,7 @@ g_desktop_app_info_equal (GAppInfo *appinfo1,
 
   if (info1->desktop_id == NULL ||
       info2->desktop_id == NULL)
-    return FALSE;
+    return info1 == info2;
 
   return strcmp (info1->desktop_id, info2->desktop_id) == 0;
 }
@@ -444,13 +452,15 @@ g_desktop_app_info_get_icon (GAppInfo *appinfo)
 }
 
 static char *
-expand_macro_single (char macro, GFile *file)
+expand_macro_single (char macro, char *uri)
 {
+  GFile *file;
   char *result = NULL;
-  char *uri, *path;
+  char *path;
 
+  file = g_file_new_for_uri (uri);
   path = g_file_get_path (file);
-  uri = g_file_get_uri (file);
+  g_object_unref (file);
   
   switch (macro)
     {
@@ -476,7 +486,6 @@ expand_macro_single (char macro, GFile *file)
     }
 
   g_free (path);
-  g_free (uri);
   
   return result;
 }
@@ -485,9 +494,9 @@ static void
 expand_macro (char              macro, 
               GString          *exec, 
               GDesktopAppInfo  *info, 
-              GList           **file_list)
+              GList           **uri_list)
 {
-  GList *files = *file_list;
+  GList *uris = *uri_list;
   char *expanded;
   
   g_return_if_fail (exec != NULL);
@@ -498,15 +507,15 @@ expand_macro (char              macro,
     case 'f':
     case 'd':
     case 'n':
-      if (files)
+      if (uris)
 	{
-	  expanded = expand_macro_single (macro, files->data);
+	  expanded = expand_macro_single (macro, uris->data);
 	  if (expanded)
 	    {
 	      g_string_append (exec, expanded);
 	      g_free (expanded);
 	    }
-	  files = files->next;
+	  uris = uris->next;
 	}
 
       break;
@@ -515,18 +524,18 @@ expand_macro (char              macro,
     case 'F':
     case 'D':
     case 'N':
-      while (files)
+      while (uris)
 	{
-	  expanded = expand_macro_single (macro, files->data);
+	  expanded = expand_macro_single (macro, uris->data);
 	  if (expanded)
 	    {
 	      g_string_append (exec, expanded);
 	      g_free (expanded);
 	    }
 	  
-	  files = files->next;
+	  uris = uris->next;
 	  
-	  if (files != NULL && expanded)
+	  if (uris != NULL && expanded)
 	    g_string_append_c (exec, ' ');
 	}
 
@@ -558,17 +567,17 @@ expand_macro (char              macro,
       break;
     }
   
-  *file_list = files;
+  *uri_list = uris;
 }
 
 static gboolean
 expand_application_parameters (GDesktopAppInfo   *info,
-			       GList            **files,
+			       GList            **uris,
 			       int               *argc,
 			       char            ***argv,
 			       GError           **error)
 {
-  GList *file_list = *files;
+  GList *uri_list = *uris;
   const char *p = info->exec;
   GString *expanded_exec = g_string_new (NULL);
   gboolean res;
@@ -584,7 +593,7 @@ expand_application_parameters (GDesktopAppInfo   *info,
     {
       if (p[0] == '%' && p[1] != '\0')
 	{
-	  expand_macro (p[1], expanded_exec, info, files);
+	  expand_macro (p[1], expanded_exec, info, uris);
 	  p++;
 	}
       else
@@ -594,11 +603,11 @@ expand_application_parameters (GDesktopAppInfo   *info,
     }
   
   /* No file substitutions */
-  if (file_list == *files && file_list != NULL)
+  if (uri_list == *uris && uri_list != NULL)
     {
       /* If there is no macro default to %f. This is also what KDE does */
       g_string_append_c (expanded_exec, ' ');
-      expand_macro ('f', expanded_exec, info, files);
+      expand_macro ('f', expanded_exec, info, uris);
     }
   
   res = g_shell_parse_argv (expanded_exec->str, argc, argv, error);
@@ -812,15 +821,17 @@ replace_env_var (char       **old_environ,
 }
 
 static GList *
-dup_list_segment (GList *start,
-		  GList *end)
+uri_list_segment_to_files (GList *start,
+			   GList *end)
 {
   GList *res;
+  GFile *file;
 
   res = NULL;
   while (start != NULL && start != end)
     {
-      res = g_list_prepend (res, start->data);
+      file = g_file_new_for_uri ((char *)start->data);
+      res = g_list_prepend (res, file);
       start = start->next;
     }
 
@@ -838,14 +849,14 @@ extern char **environ;
 #endif
 
 static gboolean
-g_desktop_app_info_launch (GAppInfo           *appinfo,
-			   GList              *files,
-			   GAppLaunchContext  *launch_context,
-			   GError            **error)
+g_desktop_app_info_launch_uris (GAppInfo           *appinfo,
+				GList              *uris,
+				GAppLaunchContext  *launch_context,
+				GError            **error)
 {
   GDesktopAppInfo *info = G_DESKTOP_APP_INFO (appinfo);
   gboolean completed = FALSE;
-  GList *old_files;
+  GList *old_uris;
   GList *launched_files;
   char **envp;
   char **argv;
@@ -860,8 +871,8 @@ g_desktop_app_info_launch (GAppInfo           *appinfo,
       
   do 
     {
-      old_files = files;
-      if (!expand_application_parameters (info, &files,
+      old_uris = uris;
+      if (!expand_application_parameters (info, &uris,
 					  &argc, &argv, error))
 	goto out;
       
@@ -875,7 +886,7 @@ g_desktop_app_info_launch (GAppInfo           *appinfo,
       sn_id = NULL;
       if (launch_context)
 	{
-	  launched_files = dup_list_segment (old_files, files);
+	  launched_files = uri_list_segment_to_files (old_uris, uris);
 	  
 	  display = g_app_launch_context_get_display (launch_context,
 						      appinfo,
@@ -909,6 +920,7 @@ g_desktop_app_info_launch (GAppInfo           *appinfo,
 
 	  g_free (display);
 	  
+	  g_list_foreach (launched_files, (GFunc)g_object_unref, NULL);
 	  g_list_free (launched_files);
 	}
       
@@ -937,7 +949,7 @@ g_desktop_app_info_launch (GAppInfo           *appinfo,
       envp = NULL;
       argv = NULL;
     }
-  while (files != NULL);
+  while (uris != NULL);
 
   completed = TRUE;
 
@@ -969,32 +981,29 @@ g_desktop_app_info_supports_files (GAppInfo *appinfo)
 }
 
 static gboolean
-g_desktop_app_info_launch_uris (GAppInfo           *appinfo,
-				GList              *uris,
-				GAppLaunchContext  *launch_context,
-				GError            **error)
+g_desktop_app_info_launch (GAppInfo           *appinfo,
+			   GList              *files,
+			   GAppLaunchContext  *launch_context,
+			   GError            **error)
 {
-  GList *files;
-  GFile *file;
+  GList *uris;
+  char *uri;
   gboolean res;
 
-  files = NULL;
-  while (uris)
+  uris = NULL;
+  while (files)
     {
-      file = g_file_new_for_uri (uris->data);
-      if (file == NULL)
-	g_warning ("Invalid uri passed to g_desktop_app_info_launch_uris");
-      
-      if (file)
-	files = g_list_prepend (files, file);
+      uri = g_file_get_uri (files->data);
+      uris = g_list_prepend (uris, uri);
+      files = files->next;
     }
   
-  files = g_list_reverse (files);
+  uris = g_list_reverse (uris);
   
-  res = g_desktop_app_info_launch (appinfo, files, launch_context, error);
+  res = g_desktop_app_info_launch_uris (appinfo, uris, launch_context, error);
   
-  g_list_foreach  (files, (GFunc)g_object_unref, NULL);
-  g_list_free (files);
+  g_list_foreach  (uris, (GFunc)g_free, NULL);
+  g_list_free (uris);
   
   return res;
 }
@@ -1087,7 +1096,7 @@ ensure_dir (DirType   type,
             GError  **error)
 {
   char *path, *display_name;
-  int err;
+  int errsv;
 
   if (type == APP_DIR)
     path = g_build_filename (g_get_user_data_dir (), "applications", NULL);
@@ -1098,16 +1107,16 @@ ensure_dir (DirType   type,
   if (g_mkdir_with_parents (path, 0700) == 0)
     return path;
 
-  err = errno;
+  errsv = errno;
   display_name = g_filename_display_name (path);
   if (type == APP_DIR)
-    g_set_error (error, G_IO_ERROR, g_io_error_from_errno (err),
+    g_set_error (error, G_IO_ERROR, g_io_error_from_errno (errsv),
                  _("Can't create user application configuration folder %s: %s"),
-                 display_name, g_strerror (err));
+                 display_name, g_strerror (errsv));
   else
-    g_set_error (error, G_IO_ERROR, g_io_error_from_errno (err),
+    g_set_error (error, G_IO_ERROR, g_io_error_from_errno (errsv),
                  _("Can't create user MIME configuration folder %s: %s"),
-                 display_name, g_strerror (err));
+                 display_name, g_strerror (errsv));
 
   g_free (display_name);
   g_free (path);
@@ -1116,12 +1125,12 @@ ensure_dir (DirType   type,
 }
 
 static gboolean
-update_default_list (const char  *desktop_id, 
-                     const char  *content_type, 
-                     gboolean     add_at_start, 
-                     gboolean     add_at_end, 
-                     gboolean     remove, 
-                     GError     **error)
+update_mimeapps_list (const char  *desktop_id, 
+		      const char  *content_type, 
+		      gboolean     add_at_start, 
+		      gboolean     add_at_end, 
+		      gboolean     remove, 
+		      GError     **error)
 {
   char *dirname, *filename;
   GKeyFile *key_file;
@@ -1139,12 +1148,12 @@ update_default_list (const char  *desktop_id,
   if (!dirname)
     return FALSE;
 
-  filename = g_build_filename (dirname, "defaults.list", NULL);
+  filename = g_build_filename (dirname, "mimeapps.list", NULL);
   g_free (dirname);
 
   key_file = g_key_file_new ();
   load_succeeded = g_key_file_load_from_file (key_file, filename, G_KEY_FILE_NONE, NULL);
-  if (!load_succeeded || !g_key_file_has_group (key_file, DEFAULT_APPLICATIONS_GROUP))
+  if (!load_succeeded || !g_key_file_has_group (key_file, ADDED_ASSOCIATIONS_GROUP))
     {
       g_key_file_free (key_file);
       key_file = g_key_file_new ();
@@ -1153,7 +1162,7 @@ update_default_list (const char  *desktop_id,
   /* Add to the right place in the list */
   
   length = 0;
-  old_list = g_key_file_get_string_list (key_file, DEFAULT_APPLICATIONS_GROUP,
+  old_list = g_key_file_get_string_list (key_file, ADDED_ASSOCIATIONS_GROUP,
 					 content_type, &length, NULL);
 
   list = g_new (char *, 1 + length + 1);
@@ -1176,7 +1185,7 @@ update_default_list (const char  *desktop_id,
   g_strfreev (old_list);
 
   g_key_file_set_string_list (key_file,
-			      DEFAULT_APPLICATIONS_GROUP,
+			      ADDED_ASSOCIATIONS_GROUP,
 			      content_type,
 			      (const char * const *)list, i);
 
@@ -1239,7 +1248,10 @@ g_desktop_app_info_set_as_default_for_type (GAppInfo    *appinfo,
 {
   GDesktopAppInfo *info = G_DESKTOP_APP_INFO (appinfo);
 
-  return update_default_list (info->desktop_id, content_type, TRUE, FALSE, FALSE, error);
+  if (!g_desktop_app_info_ensure_saved (info, error))
+    return FALSE;  
+  
+  return update_mimeapps_list (info->desktop_id, content_type, TRUE, FALSE, FALSE, error);
 }
 
 static void
@@ -1303,6 +1315,9 @@ g_desktop_app_info_set_as_default_for_extension (GAppInfo    *appinfo,
   char *dirname;
   gboolean res;
 
+  if (!g_desktop_app_info_ensure_saved (G_DESKTOP_APP_INFO (appinfo), error))
+    return FALSE;  
+  
   dirname = ensure_dir (MIMETYPE_DIR, error);
   if (!dirname)
     return FALSE;
@@ -1350,7 +1365,10 @@ g_desktop_app_info_add_supports_type (GAppInfo    *appinfo,
 {
   GDesktopAppInfo *info = G_DESKTOP_APP_INFO (appinfo);
 
-  return update_default_list (info->desktop_id, content_type, FALSE, TRUE, FALSE, error);
+  if (!g_desktop_app_info_ensure_saved (G_DESKTOP_APP_INFO (info), error))
+    return FALSE;  
+  
+  return update_mimeapps_list (info->desktop_id, content_type, FALSE, TRUE, FALSE, error);
 }
 
 static gboolean
@@ -1366,7 +1384,101 @@ g_desktop_app_info_remove_supports_type (GAppInfo    *appinfo,
 {
   GDesktopAppInfo *info = G_DESKTOP_APP_INFO (appinfo);
 
-  return update_default_list (info->desktop_id, content_type, FALSE, FALSE, TRUE, error);
+  if (!g_desktop_app_info_ensure_saved (G_DESKTOP_APP_INFO (info), error))
+    return FALSE;
+  
+  return update_mimeapps_list (info->desktop_id, content_type, FALSE, FALSE, TRUE, error);
+}
+
+static gboolean
+g_desktop_app_info_ensure_saved (GDesktopAppInfo *info,
+				 GError **error)
+{
+  GKeyFile *key_file;
+  char *dirname;
+  char *filename;
+  char *data, *desktop_id;
+  gsize data_size;
+  int fd;
+  gboolean res;
+  
+  if (info->filename != NULL)
+    return TRUE;
+
+  /* This is only used for object created with
+   * g_app_info_create_from_commandline. All other
+   * object should have a filename
+   */
+  
+  dirname = ensure_dir (APP_DIR, error);
+  if (!dirname)
+    return FALSE;
+  
+  key_file = g_key_file_new ();
+
+  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
+			 "Encoding", "UTF-8");
+  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
+			 G_KEY_FILE_DESKTOP_KEY_VERSION, "1.0");
+  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
+			 G_KEY_FILE_DESKTOP_KEY_TYPE,
+                         G_KEY_FILE_DESKTOP_TYPE_APPLICATION);
+  if (info->terminal) 
+    g_key_file_set_boolean (key_file, G_KEY_FILE_DESKTOP_GROUP,
+			    G_KEY_FILE_DESKTOP_KEY_TERMINAL, TRUE);
+
+  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
+			 G_KEY_FILE_DESKTOP_KEY_EXEC, info->exec);
+
+  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
+			 G_KEY_FILE_DESKTOP_KEY_NAME, info->name);
+
+  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
+			 G_KEY_FILE_DESKTOP_KEY_COMMENT, info->comment);
+  
+  g_key_file_set_boolean (key_file, G_KEY_FILE_DESKTOP_GROUP,
+			  G_KEY_FILE_DESKTOP_KEY_NO_DISPLAY, TRUE);
+
+  data = g_key_file_to_data (key_file, &data_size, NULL);
+  g_key_file_free (key_file);
+
+  desktop_id = g_strdup_printf ("userapp-%s-XXXXXX.desktop", info->name);
+  filename = g_build_filename (dirname, desktop_id, NULL);
+  g_free (desktop_id);
+  g_free (dirname);
+  
+  fd = g_mkstemp (filename);
+  if (fd == -1)
+    {
+      char *display_name;
+
+      display_name = g_filename_display_name (filename);
+      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+		   _("Can't create user desktop file %s"), display_name);
+      g_free (display_name);
+      g_free (filename);
+      g_free (data);
+      return FALSE;
+    }
+
+  desktop_id = g_path_get_basename (filename);
+
+  close (fd);
+  
+  res = g_file_set_contents (filename, data, data_size, error);
+  if (!res)
+    {
+      g_free (desktop_id);
+      g_free (filename);
+      return FALSE;
+    }
+
+  info->filename = filename;
+  info->desktop_id = desktop_id;
+  
+  run_update_command ("update-desktop-database", "applications");
+  
+  return TRUE;
 }
 
 /**
@@ -1386,104 +1498,42 @@ g_app_info_create_from_commandline (const char           *commandline,
 				    GAppInfoCreateFlags   flags,
 				    GError              **error)
 {
-  GKeyFile *key_file;
-  char *dirname;
   char **split;
-  char *basename, *exec, *filename, *comment;
-  char *data, *desktop_id;
-  gsize data_size;
-  int fd;
+  char *basename;
   GDesktopAppInfo *info;
-  gboolean res;
 
-  dirname = ensure_dir (APP_DIR, error);
-  if (!dirname)
-    return NULL;
+  info = g_object_new (G_TYPE_DESKTOP_APP_INFO, NULL);
+
+  info->filename = NULL;
+  info->desktop_id = NULL;
   
-  key_file = g_key_file_new ();
-
-  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
-			 "Encoding", "UTF-8");
-  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
-			 G_KEY_FILE_DESKTOP_KEY_VERSION, "1.0");
-  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
-			 G_KEY_FILE_DESKTOP_KEY_TYPE,
-                         G_KEY_FILE_DESKTOP_TYPE_APPLICATION);
-  if (flags & G_APP_INFO_CREATE_NEEDS_TERMINAL) 
-    g_key_file_set_boolean (key_file, G_KEY_FILE_DESKTOP_GROUP,
-			    G_KEY_FILE_DESKTOP_KEY_TERMINAL, TRUE);
-
-  exec = g_strconcat (commandline, " %f", NULL);
-  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
-			 G_KEY_FILE_DESKTOP_KEY_EXEC, exec);
-  g_free (exec);
-
-  /* FIXME: this should be more robust. Maybe g_shell_parse_argv and use argv[0] */
-  split = g_strsplit (commandline, " ", 2);
-  basename = g_path_get_basename (split[0]);
-  g_strfreev (split);
-  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
-			 G_KEY_FILE_DESKTOP_KEY_NAME, application_name?application_name:basename);
-
-  comment = g_strdup_printf (_("Custom definition for %s"), basename);
-  g_key_file_set_string (key_file, G_KEY_FILE_DESKTOP_GROUP,
-			 G_KEY_FILE_DESKTOP_KEY_COMMENT, comment);
-  g_free (comment);
-  
-  g_key_file_set_boolean (key_file, G_KEY_FILE_DESKTOP_GROUP,
-			  G_KEY_FILE_DESKTOP_KEY_NO_DISPLAY, TRUE);
-
-  data = g_key_file_to_data (key_file, &data_size, NULL);
-  g_key_file_free (key_file);
-
-  desktop_id = g_strdup_printf ("userapp-%s-XXXXXX.desktop", basename);
-  g_free (basename);
-  filename = g_build_filename (dirname, desktop_id, NULL);
-  g_free (desktop_id);
-  g_free (dirname);
-  
-  fd = g_mkstemp (filename);
-  if (fd == -1)
-    {
-      char *display_name;
-
-      display_name = g_filename_display_name (filename);
-      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
-		   _("Can't create user desktop file %s"), display_name);
-      g_free (display_name);
-      g_free (filename);
-      g_free (data);
-      return NULL;
-    }
-
-  desktop_id = g_path_get_basename (filename);
-
-  close (fd);
-  
-  res = g_file_set_contents (filename, data, data_size, error);
-  if (!res)
-    {
-      g_free (desktop_id);
-      g_free (filename);
-      return NULL;
-    }
-
-  run_update_command ("update-desktop-database", "applications");
-  
-  info = g_desktop_app_info_new_from_filename (filename);
-  g_free (filename);
-  if (info == NULL) 
-    g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
-		 _("Can't load just created desktop file"));
+  info->terminal = flags & G_APP_INFO_CREATE_NEEDS_TERMINAL;
+  info->startup_notify = FALSE;
+  info->hidden = FALSE;
+  if (flags & G_APP_INFO_CREATE_SUPPORTS_URIS)
+    info->exec = g_strconcat (commandline, " %u", NULL);
   else
-    info->desktop_id = g_strdup (desktop_id);
-    
-  g_free (desktop_id);
+    info->exec = g_strconcat (commandline, " %f", NULL);
+  info->nodisplay = TRUE;
+  info->binary = binary_from_exec (info->exec);
+  
+  if (application_name)
+    info->name = g_strdup (application_name);
+  else
+    {
+      /* FIXME: this should be more robust. Maybe g_shell_parse_argv and use argv[0] */
+      split = g_strsplit (commandline, " ", 2);
+      basename = g_path_get_basename (split[0]);
+      g_strfreev (split);
+      info->name = basename;
+      if (info->name == NULL)
+	info->name = g_strdup ("custom");
+    }
+  info->comment = g_strdup_printf (_("Custom definition for %s"), info->name);
   
   return G_APP_INFO (info);
 }
 
-
 static void
 g_desktop_app_info_iface_init (GAppInfoIface *iface)
 {
@@ -1607,25 +1657,68 @@ g_app_info_get_default_for_type (const char *content_type,
   return info;
 }
 
-
 /**
  * g_app_info_get_default_for_uri_scheme:
  * @uri_scheme: a string containing a URI scheme.
  *
  * Gets the default application for launching applications 
- * using this URI scheme.
- *
- * TODO: This is currently unimplemented.
+ * using this URI scheme. A URI scheme is the initial part 
+ * of the URI, up to but not including the ':', e.g. "http", 
+ * "ftp" or "sip".
  * 
- * Returns: %NULL.
+ * Returns: #GAppInfo for given @uri_scheme or %NULL on error.
  **/
 GAppInfo *
 g_app_info_get_default_for_uri_scheme (const char *uri_scheme)
 {
-  /* TODO: Implement this using giomodules, reading the gconf settings
-   * in /desktop/gnome/url-handlers
-   */
-  return NULL;
+  static gsize lookup = 0;
+  
+  if (g_once_init_enter (&lookup))
+    {
+      gsize setup_value = 1;
+      GDesktopAppInfoLookup *lookup_instance;
+      const char *use_this;
+      GIOExtensionPoint *ep;
+      GIOExtension *extension;
+      GList *l;
+
+      use_this = g_getenv ("GIO_USE_URI_ASSOCIATION");
+      
+      /* Ensure vfs in modules loaded */
+      _g_io_modules_ensure_loaded ();
+      
+      ep = g_io_extension_point_lookup (G_DESKTOP_APP_INFO_LOOKUP_EXTENSION_POINT_NAME);
+
+      lookup_instance = NULL;
+      if (use_this)
+	{
+	  extension = g_io_extension_point_get_extension_by_name (ep, use_this);
+	  if (extension)
+	    lookup_instance = g_object_new (g_io_extension_get_type (extension), NULL);
+	}
+      
+      if (lookup_instance == NULL)
+	{
+	  for (l = g_io_extension_point_get_extensions (ep); l != NULL; l = l->next)
+	    {
+	      extension = l->data;
+	      lookup_instance = g_object_new (g_io_extension_get_type (extension), NULL);
+	      if (lookup_instance != NULL)
+		break;
+	    }
+	}
+
+      if (lookup_instance != NULL)
+	setup_value = (gsize)lookup_instance;
+      
+      g_once_init_leave (&lookup, setup_value);
+    }
+
+  if (lookup == 1)
+    return NULL;
+
+  return g_desktop_app_info_lookup_get_default_for_uri_scheme (G_DESKTOP_APP_INFO_LOOKUP (lookup),
+							       uri_scheme);
 }
 
 
@@ -1746,9 +1839,11 @@ typedef struct {
   char *path;
   GHashTable *mime_info_cache_map;
   GHashTable *defaults_list_map;
-  GHashTable *defaults_list_removed_map;
+  GHashTable *mimeapps_list_added_map;
+  GHashTable *mimeapps_list_removed_map;
   time_t mime_info_cache_timestamp;
   time_t defaults_list_timestamp;
+  time_t mimeapps_list_timestamp;
 } MimeInfoCacheDir;
 
 typedef struct {
@@ -1934,10 +2029,6 @@ mime_info_cache_dir_init_defaults_list (MimeInfoCacheDir *dir)
   dir->defaults_list_map = g_hash_table_new_full (g_str_hash, g_str_equal,
 						  g_free, (GDestroyNotify)g_strfreev);
   
-  if (dir->defaults_list_removed_map != NULL)
-    g_hash_table_destroy (dir->defaults_list_removed_map);
-  dir->defaults_list_removed_map = g_hash_table_new_full (g_str_hash, g_str_equal,
-							  g_free, (GDestroyNotify)g_strfreev);
 
   key_file = g_key_file_new ();
   
@@ -1980,6 +2071,90 @@ mime_info_cache_dir_init_defaults_list (MimeInfoCacheDir *dir)
       g_strfreev (mime_types);
     }
 
+  g_key_file_free (key_file);
+  return;
+  
+ error:
+  g_free (filename);
+  g_key_file_free (key_file);
+  
+  if (mime_types != NULL)
+    g_strfreev (mime_types);
+  
+  if (load_error)
+    g_error_free (load_error);
+}
+
+static void
+mime_info_cache_dir_init_mimeapps_list (MimeInfoCacheDir *dir)
+{
+  GKeyFile *key_file;
+  GError *load_error;
+  gchar *filename, **mime_types;
+  char *unaliased_type;
+  char **desktop_file_ids;
+  int i;
+  struct stat buf;
+
+  load_error = NULL;
+  mime_types = NULL;
+
+  if (dir->mimeapps_list_added_map != NULL &&
+      !mime_info_cache_dir_out_of_date (dir, "mimeapps.list",
+					&dir->mimeapps_list_timestamp))
+    return;
+  
+  if (dir->mimeapps_list_added_map != NULL)
+    g_hash_table_destroy (dir->mimeapps_list_added_map);
+  dir->mimeapps_list_added_map = g_hash_table_new_full (g_str_hash, g_str_equal,
+							g_free, (GDestroyNotify)g_strfreev);
+  
+  if (dir->mimeapps_list_removed_map != NULL)
+    g_hash_table_destroy (dir->mimeapps_list_removed_map);
+  dir->mimeapps_list_removed_map = g_hash_table_new_full (g_str_hash, g_str_equal,
+							  g_free, (GDestroyNotify)g_strfreev);
+
+  key_file = g_key_file_new ();
+  
+  filename = g_build_filename (dir->path, "mimeapps.list", NULL);
+  if (g_stat (filename, &buf) < 0)
+    goto error;
+
+  if (dir->mimeapps_list_timestamp > 0) 
+    mime_info_cache->should_ping_mime_monitor = TRUE;
+
+  dir->mimeapps_list_timestamp = buf.st_mtime;
+
+  g_key_file_load_from_file (key_file, filename, G_KEY_FILE_NONE, &load_error);
+  g_free (filename);
+  filename = NULL;
+
+  if (load_error != NULL)
+    goto error;
+
+  mime_types = g_key_file_get_keys (key_file, ADDED_ASSOCIATIONS_GROUP,
+				    NULL, NULL);
+  if (mime_types != NULL)
+    {
+      for (i = 0; mime_types[i] != NULL; i++)
+	{
+	  desktop_file_ids = g_key_file_get_string_list (key_file,
+							 ADDED_ASSOCIATIONS_GROUP,
+							 mime_types[i],
+							 NULL,
+							 NULL);
+	  if (desktop_file_ids == NULL)
+	    continue;
+	  
+	  unaliased_type = _g_unix_content_type_unalias (mime_types[i]);
+	  g_hash_table_replace (dir->mimeapps_list_added_map,
+				unaliased_type,
+				desktop_file_ids);
+	}
+      
+      g_strfreev (mime_types);
+    }
+
   mime_types = g_key_file_get_keys (key_file, REMOVED_ASSOCIATIONS_GROUP,
 				    NULL, NULL);
   if (mime_types != NULL)
@@ -1995,7 +2170,7 @@ mime_info_cache_dir_init_defaults_list (MimeInfoCacheDir *dir)
 	    continue;
 	  
 	  unaliased_type = _g_unix_content_type_unalias (mime_types[i]);
-	  g_hash_table_replace (dir->defaults_list_removed_map,
+	  g_hash_table_replace (dir->mimeapps_list_removed_map,
 				unaliased_type,
 				desktop_file_ids);
 	}
@@ -2047,10 +2222,16 @@ mime_info_cache_dir_free (MimeInfoCacheDir *dir)
       dir->defaults_list_map = NULL;
     }
   
-  if (dir->defaults_list_removed_map != NULL)
+  if (dir->mimeapps_list_added_map != NULL)
     {
-      g_hash_table_destroy (dir->defaults_list_removed_map);
-      dir->defaults_list_removed_map = NULL;
+      g_hash_table_destroy (dir->mimeapps_list_added_map);
+      dir->mimeapps_list_added_map = NULL;
+    }
+  
+  if (dir->mimeapps_list_removed_map != NULL)
+    {
+      g_hash_table_destroy (dir->mimeapps_list_removed_map);
+      dir->mimeapps_list_removed_map = NULL;
     }
   
   g_free (dir);
@@ -2097,6 +2278,7 @@ mime_info_cache_init_dir_lists (void)
 	{
 	  mime_info_cache_dir_init (dir);
 	  mime_info_cache_dir_init_defaults_list (dir);
+	  mime_info_cache_dir_init_mimeapps_list (dir);
 	  
 	  mime_info_cache->dirs = g_list_append (mime_info_cache->dirs, dir);
 	}
@@ -2118,6 +2300,7 @@ mime_info_cache_update_dir_lists (void)
       mime_info_cache_blow_global_cache ();
       mime_info_cache_dir_init (dir);
       mime_info_cache_dir_init_defaults_list (dir);
+      mime_info_cache_dir_init_mimeapps_list (dir);
       
       tmp = tmp->next;
     }
@@ -2126,28 +2309,28 @@ mime_info_cache_update_dir_lists (void)
 static void
 mime_info_cache_init (void)
 {
-	G_LOCK (mime_info_cache);
-	if (mime_info_cache == NULL)
-	  mime_info_cache_init_dir_lists ();
-	else
-	  {
-	    time_t now;
-	    
-	    time (&now);
-	    if (now >= mime_info_cache->last_stat_time + 10)
-	      {
-		mime_info_cache_update_dir_lists ();
-		mime_info_cache->last_stat_time = now;
-	      }
-	  }
-
-	if (mime_info_cache->should_ping_mime_monitor)
-	  {
-	    /* g_idle_add (emit_mime_changed, NULL); */
-	    mime_info_cache->should_ping_mime_monitor = FALSE;
-	  }
-	
-	G_UNLOCK (mime_info_cache);
+  G_LOCK (mime_info_cache);
+  if (mime_info_cache == NULL)
+    mime_info_cache_init_dir_lists ();
+  else
+    {
+      time_t now;
+      
+      time (&now);
+      if (now >= mime_info_cache->last_stat_time + 10)
+	{
+	  mime_info_cache_update_dir_lists ();
+	  mime_info_cache->last_stat_time = now;
+	}
+    }
+  
+  if (mime_info_cache->should_ping_mime_monitor)
+    {
+      /* g_idle_add (emit_mime_changed, NULL); */
+      mime_info_cache->should_ping_mime_monitor = FALSE;
+    }
+  
+  G_UNLOCK (mime_info_cache);
 }
 
 static MimeInfoCache *
@@ -2251,13 +2434,21 @@ get_all_desktop_entries_for_mime_type (const char *base_mime_type)
 	   dir_list = dir_list->next)
 	{
 	  dir = dir_list->data;
-	  default_entries = g_hash_table_lookup (dir->defaults_list_map, mime_type);
+
+	  /* First added associations from mimeapps.list */
+	  default_entries = g_hash_table_lookup (dir->mimeapps_list_added_map, mime_type);
 	  for (j = 0; default_entries != NULL && default_entries[j] != NULL; j++)
 	    desktop_entries = append_desktop_entry (desktop_entries, default_entries[j], removed_entries);
 
-	  removed_associations = g_hash_table_lookup (dir->defaults_list_removed_map, mime_type);
+	  /* Then removed assiciations from mimeapps.list */
+	  removed_associations = g_hash_table_lookup (dir->mimeapps_list_removed_map, mime_type);
 	  for (j = 0; removed_associations != NULL && removed_associations[j] != NULL; j++)
 	    removed_entries = append_desktop_entry (removed_entries, removed_associations[j], NULL);
+
+	  /* Then system defaults (or old per-user config) (using removed associations from this dir or earlier) */
+	  default_entries = g_hash_table_lookup (dir->defaults_list_map, mime_type);
+	  for (j = 0; default_entries != NULL && default_entries[j] != NULL; j++)
+	    desktop_entries = append_desktop_entry (desktop_entries, default_entries[j], removed_entries);
 	}
 
       /* Go through all entries that support the mimetype */
@@ -2284,5 +2475,82 @@ get_all_desktop_entries_for_mime_type (const char *base_mime_type)
   return desktop_entries;
 }
 
+/* GDesktopAppInfoLookup interface: */
+
+static void g_desktop_app_info_lookup_base_init (gpointer g_class);
+static void g_desktop_app_info_lookup_class_init (gpointer g_class,
+						  gpointer class_data);
+
+GType
+g_desktop_app_info_lookup_get_type (void)
+{
+  static volatile gsize g_define_type_id__volatile = 0;
+
+  if (g_once_init_enter (&g_define_type_id__volatile))
+    {
+      const GTypeInfo desktop_app_info_lookup_info =
+      {
+        sizeof (GDesktopAppInfoLookupIface), /* class_size */
+	g_desktop_app_info_lookup_base_init,   /* base_init */
+	NULL,		/* base_finalize */
+	g_desktop_app_info_lookup_class_init,
+	NULL,		/* class_finalize */
+	NULL,		/* class_data */
+	0,
+	0,              /* n_preallocs */
+	NULL
+      };
+      GType g_define_type_id =
+	g_type_register_static (G_TYPE_INTERFACE, I_("GDesktopAppInfoLookup"),
+				&desktop_app_info_lookup_info, 0);
+
+      g_type_interface_add_prerequisite (g_define_type_id, G_TYPE_OBJECT);
+
+      g_once_init_leave (&g_define_type_id__volatile, g_define_type_id);
+    }
+
+  return g_define_type_id__volatile;
+}
+
+static void
+g_desktop_app_info_lookup_class_init (gpointer g_class,
+				      gpointer class_data)
+{
+}
+
+static void
+g_desktop_app_info_lookup_base_init (gpointer g_class)
+{
+}
+
+/**
+ * g_desktop_app_info_lookup_get_default_for_uri_scheme:
+ * @lookup: a #GDesktopAppInfoLookup
+ * @uri_scheme: a string containing a URI scheme.
+ *
+ * Gets the default application for launching applications 
+ * using this URI scheme for a particular GDesktopAppInfoLookup
+ * implementation.
+ *
+ * The GDesktopAppInfoLookup interface and this function is used
+ * to implement g_app_info_get_default_for_uri_scheme() backends
+ * in a GIO module. There is no reason for applications to use it
+ * directly. Applications should use g_app_info_get_default_for_uri_scheme().
+ * 
+ * Returns: #GAppInfo for given @uri_scheme or %NULL on error.
+ */
+GAppInfo *
+g_desktop_app_info_lookup_get_default_for_uri_scheme (GDesktopAppInfoLookup *lookup,
+						      const char            *uri_scheme)
+{
+  GDesktopAppInfoLookupIface *iface;
+  
+  g_return_val_if_fail (G_IS_DESKTOP_APP_INFO_LOOKUP (lookup), NULL);
+
+  iface = G_DESKTOP_APP_INFO_LOOKUP_GET_IFACE (lookup);
+
+  return (* iface->get_default_for_uri_scheme) (lookup, uri_scheme);
+}
+
 #define __G_DESKTOP_APP_INFO_C__
 #include "gioaliasdef.c"
