@@ -33,9 +33,20 @@
 /**
  * SECTION:gfileinputstream
  * @short_description: File input streaming operations
- * @include: gio.h
+ * @include: gio/gio.h
  * @see_also: #GInputStream, #GDataInputStream, #GSeekable
  *
+ * GFileInputStream provides input streams that take their
+ * content from a file.
+ *
+ * GFileInputStream implements #GSeekable, which allows the input 
+ * stream to jump to arbitrary positions in the file, provided the 
+ * filesystem of the file allows it. In addition to the generic 
+ * g_seekable_ API, GFileInputStream has its own API for seeking 
+ * and positioning. To find the position of a file input stream, 
+ * use g_file_input_stream_tell(). To find out if a file input 
+ * stream supports seeking, use g_file_input_stream_can_seek().
+ * To position a file input stream, use g_file_input_stream_seek().
  **/
 
 static void       g_file_input_stream_seekable_iface_init    (GSeekableIface       *iface);
@@ -105,7 +116,7 @@ g_file_input_stream_init (GFileInputStream *stream)
  * @error: a #GError location to store the error occuring, or %NULL to 
  * ignore.
  *
- * Queries a file input stream the given @attributes.his function blocks 
+ * Queries a file input stream the given @attributes. This function blocks 
  * while querying the stream. For the asynchronous (non-blocking) version 
  * of this function, see g_file_input_stream_query_info_async(). While the 
  * stream is blocked, the stream will set the pending flag internally, and 
@@ -173,8 +184,13 @@ async_ready_callback_wrapper (GObject      *source_object,
  * @callback: callback to call when the request is satisfied
  * @user_data: the data to pass to callback function
  * 
- * Queries the stream information asynchronously. For the synchronous 
- * version of this function, see g_file_input_stream_query_info(). 
+ * Queries the stream information asynchronously.
+ * When the operation is finished @callback will be called. 
+ * You can then call g_file_input_stream_query_info_finish() 
+ * to get the result of the operation.
+ *
+ * For the synchronous version of this function, 
+ * see g_file_input_stream_query_info(). 
  * 
  * If @cancellable is not %NULL, then the operation can be cancelled by
  * triggering the cancellable object from another thread. If the operation
@@ -248,15 +264,7 @@ g_file_input_stream_query_info_finish (GFileInputStream  *stream,
   return class->query_info_finish (stream, result, error);
 }
 
-/**
- * g_file_input_stream_tell:
- * @stream: a #GFileInputStream.
- * 
- * Gets the current position in the stream.
- * 
- * Returns: a #goffset with the position in the stream.
- **/
-goffset
+static goffset
 g_file_input_stream_tell (GFileInputStream *stream)
 {
   GFileInputStreamClass *class;
@@ -279,15 +287,7 @@ g_file_input_stream_seekable_tell (GSeekable *seekable)
   return g_file_input_stream_tell (G_FILE_INPUT_STREAM (seekable));
 }
 
-/**
- * g_file_input_stream_can_seek:
- * @stream: a #GFileInputStream.
- * 
- * Checks if a file input stream can be seeked.
- * 
- * Returns: %TRUE if stream can be seeked. %FALSE otherwise.
- **/
-gboolean
+static gboolean
 g_file_input_stream_can_seek (GFileInputStream *stream)
 {
   GFileInputStreamClass *class;
@@ -314,25 +314,7 @@ g_file_input_stream_seekable_can_seek (GSeekable *seekable)
   return g_file_input_stream_can_seek (G_FILE_INPUT_STREAM (seekable));
 }
 
-/**
- * g_file_input_stream_seek:
- * @stream: a #GFileInputStream.
- * @offset: a #goffset to seek.
- * @type: a #GSeekType.
- * @cancellable: optional #GCancellable object, %NULL to ignore. 
- * @error: a #GError location to store the error occuring, or 
- *     %NULL to ignore.
- * 
- * Seeks in the file input stream.
- * 
- * If @cancellable is not %NULL, then the operation can be cancelled by
- * triggering the cancellable object from another thread. If the operation
- * was cancelled, the error %G_IO_ERROR_CANCELLED will be set.
- * 
- * Returns: %TRUE if the stream was successfully seeked to the position.
- * %FALSE on error.
- **/
-gboolean
+static gboolean
 g_file_input_stream_seek (GFileInputStream  *stream,
 			  goffset            offset,
 			  GSeekType          type,
